"""Convert the parse tree created by sre_parse.parse back into a string

This library is borrowed from http://www.dalkescientific.com/Python/sre_dump.html
There are some modifications for python3 support and bug fixes.

Originally written in 2003 by Andrew Dalke <dalke@dalkescientific.com>,
Updated in 2021 by Mikihito Matsuura <me@mikit.dev>

SPDX-License-Identifier: Apache-2.0
"""
import sre_parse
from sre_constants import *
import string

__all__ = ["dump", "dump_and_offsets", "pprint", "pformat"]

_categories = {
    CATEGORY_DIGIT: "\\d",
    CATEGORY_NOT_DIGIT: "\\D",
    CATEGORY_WORD: "\\w",
    CATEGORY_NOT_WORD: "\\W",
    CATEGORY_SPACE: "\\s",
    CATEGORY_NOT_SPACE: "\\S",
    # are there more?
    }

_at_symbols = {
    AT_BEGINNING: "^",
    AT_END: "$",
    AT_BEGINNING_STRING: r"\A",
    AT_END_STRING: r"\Z",
    AT_BOUNDARY: r"\b",
    AT_NON_BOUNDARY: r"\B",
    }

# [AB\]C] to match any of A, B, ] and C
# [AB\-D] matches any one of A, B, -, and D
def _escape_in_char(i):
    c = chr(i)
    if c in ']-':
        return '\\' + c
    return c

# [--A] means from "-" to "A"
def _escape_in_range(i):
    c = chr(i)
    if c in ']':
        return '\\' + c
    return c

def _dump_in(av):
    # XXX is this proper escaping?
    words = []
    emit = words.append
    first_flg = 1
    for x in av:
        if x[0] == NEGATE:
            if first_flg == 1:
                emit("^")
            else:
                raise AssertionError("negate not first?")
        elif x[0] == LITERAL:
            if first_flg and x[1] == ord("^"):
                emit("\\^")
            else:
                emit(_escape_in_char(x[1]))
        elif x[0] == RANGE:
            emit(_escape_in_range(x[1][0]) + "-" +
                 _escape_in_range(x[1][1]))
        elif x[0] == CATEGORY:
            emit(_categories[x[1]])
        else:
            raise NotImplementedError(x[0])
        first_flg = 0
    return "[" + "".join(words) + "]"
        

# I could use sre.escape but I don't like that it
# escapes "<", ">", and some other characters
_escape_table = {"\000": "\\000"}  # special-case 0
for i in range(256):
    _escape_table[i] = "\\" + chr(i)
for c in string.ascii_letters + string.digits + "<>: /":
    _escape_table[ord(c)] = c
del c, i


def _dump(pattern, groupdict):
    pos = [0]
    offsets = []
    words = []
    def emit(s):
        words.append(s)
        pos[0] = pos[0] + len(s)

    def include(suboffsets):
        x = pos[0]
        for e, i, j in suboffsets:
            offsets.append( (e, i+x, j+x) )

    start = pos[0]
    
    for term in pattern:
        op, av = term
        
        if op == LITERAL:
            emit(_escape_table[av])
        elif op == NOT_LITERAL:
            emit("[^%s]" % chr(av))
        elif op == IN:
            emit(_dump_in(av))
        elif op in (MAX_REPEAT, MIN_REPEAT):
            i, j, subtree = av
            s, suboffsets = _dump(subtree, groupdict)
            include(suboffsets)
            emit(s)
            if j == MAXREPEAT:
                if i == 0:
                    emit("*")
                elif i == 1:
                    emit("+")
                else:
                    emit("{%d,}" % i)
            else:
                if i == 0 and j == 1:
                    emit("?")
                elif i == j:
                    emit("{%d}" % i)
                elif i == 0:
                    emit("{,%d}" % j)
                else:
                    emit("{%d,%d}" % (i, j))
            if op == MIN_REPEAT:
                emit("?")
        elif op == SUBPATTERN:
            groupnum, _, _, subtree = av
            s, suboffsets = _dump(subtree, groupdict)
            if groupnum in groupdict:
                name = groupdict[groupnum]
                emit("(?P<%s>" % (name,))  #")  emacs cruft
            elif groupnum is None:
                emit("(?:")  #")  emacs cruft
            else:
                emit("(")
            include(suboffsets) 
            emit("%s)" % s)
            
        elif op == ASSERT:
            dir, subtree = av
            dir = {1:"", -1:"<"}[dir]
            s, suboffsets = _dump(subtree, groupdict)
            emit("(?%s=" % dir)  #")
            include(suboffsets)
            emit("%s)" % s)
        elif op == ASSERT_NOT:
            dir, subtree = av
            dir = {1:"", -1:"<"}[dir]
            s, suboffsets = _dump(subtree, groupdict)
            emit("(?%s!" % dir)
            include(suboffsets)
            emit("%s)" % s)
                       
        elif op == BRANCH:
            # av[0] is always None
            emit("(?:")
            for x in av[1][:-1]:
                s, suboffsets = _dump(x, groupdict)
                include(suboffsets)
                emit(s)
                emit("|")
            s, suboffsets = _dump(av[1][-1], groupdict)
            include(suboffsets)
            emit(s)
            emit(")")
        elif op == ANY:
            emit(".")
        elif op == AT:
            emit(_at_symbols[av])
        elif op == GROUPREF:
            if av in groupdict:
                emit("(?P=%s)" % groupdict[av])
            else:
                emit("\\%d" % av)
        else:
            raise NotImplementedError(op)

        # These are added in creation order, inner before outer.
        offsets.append( (term, start, pos[0]) )
        start = pos[0]

    return "".join(words), offsets

class NoGroupNames:
    def __getitem__(self, i):
        assert i < 100,  "that would be silly"
        return "group_%d" % i
    def __contains__(self, i):
        return 0

def dump_and_offsets(tree):
    # You really should pass in a SubTree, but I'll let you
    # get away with a couple common variations
    if isinstance(tree, list):
        # used when building a list by hand; ignore group names
        d = NoGroupNames()
    elif isinstance(tree, tuple) and len(tree) == 2:
        # an (op, av) 2-pl element in the tree
        tree = [tree]
        d = NoGroupNames()
    else:
        if isinstance(tree, str):
            # pass in a string and I'll convert it for you
            tree = sre_parse.parse(tree)
        
        # need the reverse mapping to get the names
        d = {}
        for k, v in list(tree.state.groupdict.items()):
            d[v] = k

    return _dump(tree, d)

def dump(tree):
    return dump_and_offsets(tree)[0]


def pprint(tree, stream=None):
    s, offsets = dump_and_offsets(tree)
    print(s, file=stream)
    for expr, i, j in offsets:
        print(" "*i + "-"*(j-i) + " " *(len(s)-j+1) + s[i:j], file=stream)

def pformat(tree):
    import io as StringIO
    sio = StringIO.StringIO()
    pprint(tree, sio)
    return sio.getvalue()

def _do_test(input, expected = None):
    if expected is None:
        expected = input
    tree = sre_parse.parse(input)
    try:
        output, offsets = dump_and_offsets(tree)
    except (NotImplementedError, KeyError) as msg:
        import traceback
        print("   ===== Cannot compute =====")
        print("Input was", repr(input))
        print("Exception mesage:", msg)
        traceback.print_exc()
        return False
        
    if output != expected:
        print("   ===== Different =====")
        print("Input was", repr(input))
        print("Expected:", repr(expected))
        print("Output is", repr(output))
        print(tree)
        return False

    # Check the offsets are generated correctly
    for expr, i, j in offsets:
        a = output[i:j]
        # Use the SubPattern to get the pattern names correct.
        b = dump(sre_parse.SubPattern(tree.pattern, [expr]))
        if a != b:
            print("   ===== bad substrings =====")
            print("Input was", repr(input))
            print("Excepcted", repr(a))
            print("Redump is", repr(b))
            return False
    return True
                       

def test():
    # Some of these taken from 'perlre', some from AMK's docs
    # and many I just made up
    test_values = r"""
A AB ABC [^A] [^ABC] A+ A* A? AA* A{2} A{2,4} A{9,12} A{2,} A{,9}
[\d] [^\d] [A-Za-z0-9_]+ . .*
AB|CD AB|CD|EF|GH (AB|CD)*
(a??) a*? a{3,}? ab{4,7}?
(?P<test>ABC*) (?P<a>x)|(?P<b>y)
[b:]+ (b)|(:+) a|(b)
(?:Q)(Q) ^A*$
(?:(?P<a1>a)|(?P<b2>b))(?P<c3>c)?
(?=AB)C (?!CD)DC AB(?<=CD) AB(?<!CD)
[ABC]+(?=D).*$ <.*?>
(?P<name>[a-zA-Z]+)(?P=name)
[AB\]C] [--A] [ABC\-D] [\^ABC]
""".split()
    passed = True
    for input in test_values:
        passed = passed and _do_test(input)

    # sre_parse does some optimization so I
    # have to allow variant forms
    special_values = r"""
      \d [\d]
      \d+ [\d]+
      \w [\w]
      \W [\W]
      \s [\s]
      \S [\S]
      A|B [AB]
      A|B|C|D [ABCD]
      (?:a|b|c) (?:[abc])
      .*\..*$ .*\..*$
      .*\.([^b]..|.[^a].|..[^t])$ .*\.([^b]..|.[^a].|..[^t])$
      .*\.([^b].?.?|.[^a]?.?|..?[^t]?) .*\.([^b].?.?|.[^a]?.?|..?[^t]?)
      .*\.(?!bat$).*$ .*\.(?!bat$).*$
      .*\.(?!bat$|exe$) .*\.(?!bat$|exe$)
      """.split("\n")[1:-1]
    for x in special_values:
        if not x.strip():
            continue
        input, expected = x.split()
            
        
        passed = passed and _do_test(input, expected)

    # issues with spaces
    others = [
        r"^([^ ]*) *([^ ]*)",
        r"(.)\1",
        r"/Time: (..):(..):(..)/",
        r"\Aabc",
        r"abc\b",
        r"abc\Bqwe",
        r"qwerty\Z",
        (r"\n", "\\\n"),
        ]
    for x in others:
        if isinstance(x, str):
            passed = passed and _do_test(x)
        else:
            input, output = x
            passed = passed and _do_test(input, output)


    s = pformat("(AB*|C?D){2,3}((?P<name>\d+)(?:a|b))+")

    if passed:
        print("All tests passed.")
    else:
        print("Test failure.")
        sys.exit(1)

if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        test()
    else:
        in_s = sys.argv[1]
        x = sre_parse.parse(in_s)
        out_s = dump(x)
        print("Input :", repr(in_s))
        print("Output:", repr(out_s))




from . import mlr
from .mlr import Node, NodeList
from typing import Callable, Union, Optional
import inspect

class Transformer:
  def __init__(self, sequenceRule: Optional[Callable[[NodeList], Node]] = None, nodeRules: dict[Node, Callable[[Node], Node]] = {}):
    self._nodeRules = nodeRules
    self._sequenceRule = sequenceRule

  def __call__(self, regex: Union[Node, NodeList, str]):
    if isinstance(regex, Node):
      self.transform_node(regex)
      return regex
    if isinstance(regex, NodeList):
      self.transform_nodes(regex)
      return regex
    if isinstance(regex, str):
      node = mlr.parse(regex)
      self.transform_nodes(node)
      return node.dump()
    raise NotImplementedError()

  def transform_node(self, node: Node):
    rule = self._nodeRules.get(type(node), None)
    if rule is not None:
      rule(node)
      return
    node._traverse(self)

  def transform_nodes(self, nodes: NodeList):
    if self._sequenceRule is not None:
      self._sequenceRule(nodes)
      return
    nodes._traverse(self)

  def add_rule(self, target: type[Union[Node, NodeList]]=None):
    def f(rule):
      nonlocal target
      if target is None:
        spec = inspect.getfullargspec(rule)
        target = spec.annotations[spec.args[0]]
      if issubclass(target, NodeList):
        self._sequenceRule = rule
      elif issubclass(target, Node):
        self._nodeRules[target] = rule
      else:
        raise TypeError()
    return f



import abc
import collections.abc
import inspect
from typing import Optional
import sre_parse as _sre_parse
from . import _sre_dump

class Node(abc.ABC):
  @classmethod
  @property
  @abc.abstractmethod
  def _sre_id(cls) -> str:
    pass

  @abc.abstractmethod
  def _to_sre(self):
    pass

  def _to_sre_tuple(self):
    return (self._sre_id, self._to_sre())

  @classmethod
  @abc.abstractmethod
  def _from_sre(cls, *args, **kwargs):
    pass

  def _traverse(self, handler):
    pass

  def dump(self):
    return _sre_dump.dump([self._to_sre_tuple()])

  def __repr__(self):
    t = self._to_sre()
    if not isinstance(t, tuple):
      t = (t,)
    return f"{type(self).__name__}{t}"


class NodeList:
  _nodes: dict[int, type[Node]] = dict()

  def __init__(self, *nodes: list[Node]):
    self.nodes = nodes

  @abc.abstractmethod
  def _to_sre(self) -> list[tuple]:
    return list(map(lambda node: node._to_sre_tuple(), self.nodes))

  @classmethod
  def _from_sre(cls, ops):
    def decode(c, args):
      factory = cls._nodes[c]._from_sre
      if len(inspect.signature(factory).parameters) == 1:
        return factory(args)
      else:
        return factory(*args)
    return NodeList(*[decode(*op) for op in ops])

  def _traverse(self, handler):
    for child in self.nodes:
      handler(child)

  def dump(self):
    return _sre_dump.dump(self._to_sre())

  @classmethod
  def _register_type(cls, node_type: type[Node]):
    cls._nodes[node_type._sre_id] = node_type
    return node_type

  def __repr__(self):
    return f"[{', '.join(map(repr, self.nodes))}]"


srenode = NodeList._register_type


@srenode
class Literal(Node):
  from sre_constants import LITERAL as _sre_id

  def __init__(self, code: int):
    self.code = code

  @property
  def string(self):
    return chr(self.code)

  @string.setter
  def string(self, x):
    self.code = ord(x)

  def _to_sre(self):
    return self.code

  @classmethod
  def _from_sre(cls, code: int):
    return cls(code)


@srenode
class NotLiteral(Node):
  from sre_constants import NOT_LITERAL as _sre_id

  def __init__(self, code: int):
    self.code = code

  @property
  def string(self):
    return chr(self.code)

  @string.setter
  def string(self, x):
    self.code = ord(x)

  def _to_sre(self):
    return self.code

  @classmethod
  def _from_sre(cls, code: int):
    return cls(code)


@srenode
class In(Node):
  from sre_constants import IN as _sre_id

  def __init__(self, target: NodeList):
    self.target = target

  def _to_sre(self):
    return self.target._to_sre()

  @classmethod
  def _from_sre(cls, ops: list):
    return cls(NodeList._from_sre(ops))

  def _traverse(self, handler):
    self.target._traverse(handler)


@srenode
class Negate(Node):
  from sre_constants import NEGATE as _sre_id

  def __init__(self):
    pass

  def _to_sre(self):
    return None

  @classmethod
  def _from_sre(cls, _=None):
    return cls()


@srenode
class Range(Node):
  from sre_constants import RANGE as _sre_id

  def __init__(self, lo_code: int, hi_code: int):
    self.lo_code = lo_code
    self.hi_code = hi_code

  @property
  def lo_string(self):
    return chr(self.lo_code)

  @property
  def hi_string(self):
    return chr(self.hi_code)

  @lo_string.setter
  def lo_string(self, x):
    self.lo_code = ord(x)

  @hi_string.setter
  def hi_string(self, x):
    self.hi_code = ord(x)

  def _to_sre(self):
    return self.lo_code, self.hi_code

  @classmethod
  def _from_sre(cls, lo_code, hi_code):
    return cls(lo_code, hi_code)


@srenode
class Category(Node):
  from sre_constants import CATEGORY as _sre_id, CHCODES

  _categories = {item.name.lower(): item for item in CHCODES}

  def __init__(self, category: str):
    self.category = category

  def _to_sre(self):
    return self._categories[self.category]

  @classmethod
  def _from_sre(cls, category):
    return cls(category.name.lower())


@srenode
class At(Node):
  from sre_constants import AT as _sre_id, ATCODES

  _positions = {item.name.lower(): item for item in ATCODES}

  def __init__(self, position: str):
    self.position = position

  def _to_sre(self):
    return self._positions[self.position]

  @classmethod
  def _from_sre(cls, position):
    return cls(position.name.lower())


@srenode
class MinRepeat(Node):
  from sre_constants import MIN_REPEAT as _sre_id
  from sre_constants import MAXREPEAT

  def __init__(self, min: Optional[int], max: Optional[int], target: NodeList):
    self.min = min
    self.max = max
    self.target = target

  def _to_sre(self):
    max = self.MAXREPEAT if self.max is None else self.max
    return (self.min, max, self.target._to_sre())

  @classmethod
  def _from_sre(cls, min, max, target):
    max = None if max == cls.MAXREPEAT else max
    return cls(min, max, NodeList._from_sre(target))

  def _traverse(self, handler):
    self.target._traverse(handler)


@srenode
class MaxRepeat(Node):
  from sre_constants import MAX_REPEAT as _sre_id
  from sre_constants import MAXREPEAT

  def __init__(self, min: Optional[int], max: Optional[int], target: NodeList):
    self.min = min
    self.max = max
    self.target = target

  def _to_sre(self):
    max = self.MAXREPEAT if self.max is None else self.max
    return (self.min, max, self.target._to_sre())

  @classmethod
  def _from_sre(cls, min, max, target):
    max = None if max == cls.MAXREPEAT else max
    return cls(min, max, NodeList._from_sre(target))

  def _traverse(self, handler):
    self.target._traverse(handler)


@srenode
class SubPattern(Node):
  from sre_constants import SUBPATTERN as _sre_id

  def __init__(self, gid: int, target: NodeList):
    self.gid = gid
    self.target = target

  def _to_sre(self):
    return (self.gid, 0, 0, self.target._to_sre())

  @classmethod
  def _from_sre(cls, gid, _unk1, _unk2, target):
    return cls(gid, NodeList._from_sre(target))

  def _traverse(self, handler):
    self.target._traverse(handler)


@srenode
class Branch(Node):
  from sre_constants import BRANCH as _sre_id

  def __init__(self, targets: list[NodeList]):
    self.targets = targets

  def _to_sre(self):
    return (None, list(map(NodeList._to_sre, self.targets)))

  @classmethod
  def _from_sre(cls, _unk1, targets):
    return cls(list(map(NodeList._from_sre, targets)))

  def _traverse(self, handler):
    for target in self.targets:
      target._traverse(handler)


@srenode
class Any(Node):
  from sre_constants import ANY as _sre_id

  def __init__(self):
    pass

  def _to_sre(self):
    return None

  @classmethod
  def _from_sre(cls, _unk1):
    return cls()


def parse(pattern):
  return NodeList._from_sre(_sre_parse.parse(pattern))
